

package com.hello2morrow.sonargraph.languageprovider.cplusplus.controller.settings

import java.nio.charset.StandardCharsets
import java.util.regex.*

import org.slf4j.Logger
import org.slf4j.LoggerFactory

import com.hello2morrow.foundation.activity.IWorkerContext
import com.hello2morrow.foundation.file.FileUtility
import com.hello2morrow.foundation.file.TDirectoryFileFilter
import com.hello2morrow.foundation.file.TrueZipFacade
import com.hello2morrow.foundation.utilities.ExceptionUtility
import com.hello2morrow.foundation.utilities.Platform
import com.hello2morrow.foundation.utilities.StringUtility
import com.hello2morrow.foundation.utilities.TemplateUtility
import com.hello2morrow.foundation.windows.VisualStudioInstallationsDetector
import com.hello2morrow.sonargraph.languageprovider.cplusplus.model.settings.CompilerDefinitionSearchPath

import de.schlichtherle.truezip.file.TFile

final class VisualCppConfigurator implements IVisualCppConfigurator
{
    private static final Logger LOGGER = LoggerFactory.getLogger(VisualCppConfigurator.class)

    public static final String ARCHITECTURE = "arch"
    public static final String INSTALLATION_DIR = "instDir"

    private static final String OPTSPY = "optspy"
    private static final String OPTSPY_C = OPTSPY + ".c"
    private static final String OPTSPY_EXE = OPTSPY + ".exe"
    private static final String FAILED_SUB_DIRECTORY_PATH = "failed"
    private static final String X86 = "x86"

    private String m_compilerDefinitionTemplate = "compilerDefinition/VisualCpp.cdef";

    private static final String PROPERTY_DESCRIPTION = "description"
    private static final String PROPERTY_OPTIONS = "Options"
    private final TFile m_configDirectory
    private final TFile m_failedConfigDirectory
    private final Map<String, String> m_locations = new LinkedHashMap<>();
    private String m_latestVersion

    public VisualCppConfigurator(final TFile configDirectory)
    {
        assert Platform.isWindows()
        assert configDirectory != null
        assert configDirectory.isDirectory()

        m_configDirectory = configDirectory
        m_failedConfigDirectory = new TFile(m_configDirectory, FAILED_SUB_DIRECTORY_PATH)

        List<VisualStudioInstallationsDetector.Location> locations = VisualStudioInstallationsDetector.getLocations();
        for (VisualStudioInstallationsDetector.Location next : locations)
        {
            m_locations.put(next.getVersion(), next.getFile().getNormalizedAbsolutePath());
        }

        if (m_locations.isEmpty())
        {
            LOGGER.error("Failed to analyze Windows Registry to detect Visual Studio Compiler installations")
            return
        }

        m_latestVersion = VisualStudioInstallationsDetector.getLatestLocation(locations).orElse(null);
    }

    @Override
    public TFile getConfigurationDirectory()
    {
        return m_configDirectory
    }

    private TFile writeOptSpySource(TFile dir)
    {
        InputStream s = getClass().getClassLoader().getResourceAsStream(OPTSPY_C)
        TFile cSourceFile = new TFile(dir, OPTSPY_C)

        if (cSourceFile.exists())
        {
            cSourceFile.rm()
        }

        cSourceFile << s
        s.close()
        return cSourceFile
    }

    private String nextToken(StringReader reader)
    {
        def writer = new StringWriter()
        int c = reader.read()

        while (c > 0 && Character.isWhitespace(c))
        {
            c = reader.read()
        }
        if (c < 0)
        {
            return null
        }
        if (c == '"')
        {
            c = reader.read()
            while (c != '"')
            {
                assert c > 0
                writer.write(c)
                c = reader.read()
            }
            return writer.toString()
        }
        writer.write(c)
        c = reader.read()
        while (c > 0 && c != ' ')
        {
            writer.write(c)
            c = reader.read()
        }
        return writer.toString()
    }

    private List parseOptions(Map props, String val)
    {
        def opts = new ArrayList()
        def sawinc = false
        def reader = new StringReader(val)
        String token

        while ((token = nextToken(reader)) != null)
        {
            if (token.startsWith("-D"))
            {
                opts.add(token)
                if (token.startsWith("-D_MSC_VER="))
                {
                    props["version"] = token.substring(11)
                }
            }
            else if (token == "-I")
            {
                opts.add(nextToken(reader))
            }
            else if (token.startsWith("-I"))
            {
                opts.add(token.substring(2))
            }
        }
        return opts
    }

    private Map analyzeCompilerOutput(String output)
    {
        def propMap = new HashMap()
        List options

        output.eachLine
        {
            if (it.contains("Microsoft (R)"))
            {
                propMap[PROPERTY_DESCRIPTION] = it
            }
            else
            {
                def eqPos = it.indexOf('=')

                if (eqPos > 0)
                {
                    def key = it.substring(0, eqPos)
                    def val = it.substring(eqPos+1)

                    if (key != "MSC_CMD_FLAGS")
                    {
                        propMap[key] = val
                    }
                    else
                    {
                        propMap[PROPERTY_OPTIONS] = parseOptions(propMap, val)
                    }
                }
            }
        }

        if (propMap[PROPERTY_DESCRIPTION] == null)
        {
            propMap[PROPERTY_DESCRIPTION] = ""
        }

        if (propMap[PROPERTY_OPTIONS] == null)
        {
            propMap[PROPERTY_OPTIONS] = new ArrayList(0);
        }
        return propMap
    }

    private TFile generateConfigurationFor(TFile dir, TFile instDir, TFile binDir, String version, String arch)
    {
        String fileName = generateFileName(version, arch)
        LOGGER.info("Generate C++ Compiler Definition '$fileName' in directory '${m_configDirectory.normalizedAbsolutePath}'")

        TFile outFile = new TFile(m_configDirectory, fileName)
        dir.mkdir()

        TFile configProperties = new TFile(dir, "config.properties")
        if (configProperties.exists())
        {
            configProperties.rm()
        }
        def instDirStr = instDir.path.replace("\\", "\\\\")

        configProperties << "instDir=$instDirStr" + StringUtility.LINE_SEPARATOR + "arch=$arch" + StringUtility.LINE_SEPARATOR

        def baseCommand = new ArrayList<>()

        String shellPath = getShellPath(instDir, version);
        if(shellPath == null)
        {
            LOGGER.error("Failed to determine path of '" + VisualStudioInstallationsDetector.VS_COMMAND_PROMPT_SCRIPT
                    + "' or '" + VisualStudioInstallationsDetector.VS_DEVELOPER_COMMAND_PROMPT_SCRIPT + "' in $binDir");
            return null;
        }

        baseCommand.add(shellPath)
        baseCommand.add(arch)
        baseCommand.add("&&")
        baseCommand.add("cl");

        def createExecutableCommand = new ArrayList<>(baseCommand)
        createExecutableCommand.add("/MD")
        createExecutableCommand.add("-Fo" + new TFile(dir,  OPTSPY + ".obj").getNormalizedAbsolutePath())
        createExecutableCommand.add("-Fe" + new TFile(dir, OPTSPY_EXE).getNormalizedAbsolutePath())
        createExecutableCommand.add(new TFile(m_configDirectory, OPTSPY_C).getNormalizedAbsolutePath())

        LOGGER.debug("Executing: {}", createExecutableCommand)
        Process proc = createExecutableCommand.execute()
        def writer = new StringWriter()

        writer << proc.err
        writer << proc.in

        int result = proc.waitFor()
        TFile objFile = new TFile(dir, "optspy.obj")

        if (objFile.exists())
        {
            objFile.rm()
        }
        if (result != 0)
        {
            String msg = writer.toString()
            LOGGER.error("Cannot compile or link optspy: $msg")
            generateFailedFile(fileName, "Cannot compile or link optspy: $msg")
            return null
        }

        def createCOptionsCommand = new ArrayList<>(baseCommand)
        createCOptionsCommand.add("-B1" + new TFile(dir, OPTSPY_EXE).getNormalizedAbsolutePath())
        createCOptionsCommand.add("x.c")

        LOGGER.debug("Executing: {}", createCOptionsCommand)
        proc = createCOptionsCommand.execute()
        writer = new StringWriter()

        writer << proc.err
        writer << proc.in
        result = proc.waitFor()
        String output = writer.toString();
        if (result < 0)
        {
            LOGGER.error("Failed to execute command for analyzing C options for $version, $arch: $output")
        }
        else
        {
            LOGGER.debug("C compiler output: $output");
        }
        def cProperties = analyzeCompilerOutput(output)
        List cOptions = cProperties[PROPERTY_OPTIONS]
        if (cOptions.isEmpty())
        {
            LOGGER.warn("No C options determined for compiler: architecture '$arch', version '$version'. \nCompiler output: $output")
        }

        def createCppOptionsCommand = new ArrayList<>(baseCommand)
        createCppOptionsCommand.add("-Bx" + new TFile(dir, OPTSPY_EXE).getNormalizedAbsolutePath())
        createCppOptionsCommand.add("x.cpp")

        LOGGER.debug("Executing: {}", createCppOptionsCommand)
        proc = createCppOptionsCommand.execute()

        writer = new StringWriter()
        writer << proc.err
        writer << proc.in
        result = proc.waitFor()
        output = writer.toString();
        if (result < 0)
        {
            LOGGER.error("Failed to execute command for analyzing C++ options: $output")
        }
        else
        {
            LOGGER.debug("C++ compiler output: $output");
        }
        def cppProperties = analyzeCompilerOutput(output)
        List cppOptions = cppProperties[PROPERTY_OPTIONS]
        if (cppOptions.isEmpty())
        {
            LOGGER.warn("No C++ options determined for compiler: architecture '$arch', version '$version'. \nCompiler output: $output")
        }

        List sharedOptions = new ArrayList()

        for (String opt : cOptions)
        {
            if (cppOptions.contains(opt))
            {
                sharedOptions.add(opt)
            }
        }
        cOptions.removeAll(sharedOptions)
        cppOptions.removeAll(sharedOptions)

        def props = cProperties

        props.remove(PROPERTY_OPTIONS)
        props["sharedOptions"] = sharedOptions
        props["cOptions"] = cOptions
        props["cppOptions"] = cppOptions

        def fixedProps = new ArrayList()
        final String vcvarsallPath = getShellPath(instDir, version);
        if (vcvarsallPath != null)
        {
            String vcvarsallDir = new TFile(vcvarsallPath).getParent();
            fixedProps.add("$INSTALLATION_DIR=$vcvarsallDir")
        }
        else
        {
            LOGGER.error("Failed to determine directory of vcvarsall.bat");
        }

        fixedProps.add("$ARCHITECTURE=$arch")

        props["fixedProperties"] = fixedProps

        InputStream templateStream = getClass().getClassLoader().getResourceAsStream(m_compilerDefinitionTemplate)

        writer = new StringWriter()
        writer << templateStream
        templateStream.close()

        try
        {
            outFile << TemplateUtility.expandTemplateText(writer.toString(), props)
        }
        catch (Throwable t)
        {
            LOGGER.error("Fail to expand C++ Visual Studio compiler definition groovy template: {}", ExceptionUtility.collectFirstAndLast(t));
            if (outFile.exists())
            {
                try
                {
                    outFile.rm();
                }
                catch (IOException ex)
                {
                    LOGGER.error("Failed to delete compiler definition file '" + outFile.getAbsolutePath() + "': {}", ExceptionUtility.collectFirstAndLast(ex));
                }
            }
            return null;
        }

        return outFile
    }

    private String getShellPath(TFile instDir, String version)
    {
        assert instDir != null : "Parameter 'instDir' of method getShellPath() must not be null";

        //TODO [IK] For VisualStudio 2017 there are additional batch files in C:\Program Files (x86)\Microsoft Visual Studio 2017\Community\VC\Auxiliary\Build
        // for other architectures: vcvars64.bat, vcvarsamd64_x86.bat, vcvarsx86_amd64.bat

        //<C:\Program Files (x86)\Microsoft Visual Studio 2017\Community>\VC\Tools\MSVC\<14.10.25017>\bin\HostX64\x64\cl.exe

        if (VisualStudioInstallationsDetector.is2017VersionOrNewer(version))
        {
            TFile standardShell = new TFile(new TFile(instDir, "VC/Auxiliary/Build"), VisualStudioInstallationsDetector.VS_COMMAND_PROMPT_SCRIPT);
            if (standardShell.exists())
            {
                return standardShell.getNormalizedAbsolutePath();
            }
            return null;
        }

        TFile standardShell = new TFile(instDir, VisualStudioInstallationsDetector.VS_COMMAND_PROMPT_SCRIPT);
        if (standardShell.exists())
        {
            return standardShell.getNormalizedAbsolutePath();
        }
        return null;
    }

    private String generateFileName(String version, String arch)
    {
        return generateDefinitionName(version, arch) + ".cdef"
    }

    private String generateDefinitionName(String version, String arch)
    {
        return "VisualCpp_${version}_${arch}"
    }

    @Override
    public List<TFile> generateConfigurations(IWorkerContext workerContext)
    {
        List<TFile> generatedDefinitionFiles = new ArrayList<>()
        m_latestVersion = "0.0"
        if (m_locations == null)
        {
            return generatedDefinitionFiles
        }

        double highestVersion = 0.0

        TFile optSpySource = writeOptSpySource(m_configDirectory)

        for (Map.Entry<String, String> nextLocation : m_locations.entrySet())
        {
            String version = nextLocation.getKey();
            if (version.startsWith("Framework"))
            {
                continue
            }
            if (VisualStudioInstallationsDetector.is2017VersionOrNewer(version))
            {
                List<TFile> generatedVs2017Files = generateConfigurationsFor2017AndNewer(version, nextLocation.getValue(), workerContext);
                if (generatedVs2017Files.size() > 0)
                {
                    double doubleVersion = Double.valueOf(version)
                    if (doubleVersion > highestVersion)
                    {
                        m_latestVersion = version
                        highestVersion = doubleVersion;
                    }
                }
                generatedDefinitionFiles.addAll(generatedVs2017Files);
                continue;
            }
            String fileName = generateFileName(version, X86)
            if (configFileExists(fileName))
            {
                LOGGER.info("Using previously generated C++ Compiler Definition '${fileName}' in directory '${m_configDirectory.normalizedAbsolutePath}'")
                generatedDefinitionFiles.add(new TFile(m_configDirectory, fileName))
                if (VisualStudioInstallationsDetector.isConvertibleVersion(version))
                {
                    double doubleVersion = Double.valueOf(version)
                    if (doubleVersion > highestVersion)
                    {
                        m_latestVersion = version
                        highestVersion = doubleVersion;
                    }
                }
            }
            else if (failedConfigFileExists(fileName))
            {
                LOGGER.info("Skipping generation of previously failed C++ Compiler Definition '" + fileName + "'")
            }
            else
            {
                workerContext.working("Generate definition for Visual Studio C++ Compiler '${version}' '${X86}'", true);
                
                TFile instDir = new TFile(new TFile(nextLocation.getValue()), "vc");
                TFile binDir = new TFile(instDir, "bin");
                TFile dir = new TFile(m_configDirectory, generateDefinitionName(version, X86))
                try
                {
                    TFile definitionFile = generateConfigurationFor(dir, instDir, binDir, version, X86)
                    if (definitionFile != null)
                    {
                        generatedDefinitionFiles.add(definitionFile)
                        if (VisualStudioInstallationsDetector.isConvertibleVersion(version))
                        {
                            double doubleVersion = Double.valueOf(version)
                            if (doubleVersion > highestVersion)
                            {
                                m_latestVersion = version
                                highestVersion = doubleVersion;
                            }
                        }
                    }
                }
                catch (Throwable t)
                {
                    LOGGER.error("Failed to generate definition for Visual Studio C++ Compiler '${version}' '${X86}' ", t);
                    generateFailedFile(generateFileName(version, X86), ExceptionUtility.collectFirstAndLast(t));
                }
            }

            //generate definitions for other architectures than x86
            TFile instDir = new TFile(new TFile(nextLocation.getValue()), "vc")
            TFile binDir = new TFile(instDir, "bin")

            for (TFile nextDir : binDir.listFiles(new TDirectoryFileFilter()))
            {
                TFile cl = new TFile(nextDir, "cl.exe")
                if (cl.exists())
                {
                    fileName = generateFileName(version, nextDir.getName())
                    if (configFileExists(fileName))
                    {
                        generatedDefinitionFiles.add(new TFile(m_configDirectory, fileName))
                        LOGGER.info("Using previously generated C++ Compiler Definition '${fileName}' in directory '${m_configDirectory.normalizedAbsolutePath}'")
                    }
                    else if (failedConfigFileExists(fileName))
                    {
                        LOGGER.info("Skipping generation of previously failed C++ Compiler Definition '" + fileName + "'")
                    }
                    else
                    {
                        def arch = nextDir.getName()
                        workerContext.working("Generate definition for Visual Studio C++ Compiler '${version}' '${arch}'", true)
                        TFile definitionFile = generateConfigurationFor(new TFile(m_configDirectory, generateDefinitionName(version, arch)), instDir, nextDir, version, arch)
                        if (definitionFile != null)
                        {
                            generatedDefinitionFiles.add(definitionFile)
                        }
                    }
                }
            }
        }
        try
        {
            optSpySource.rm()
        }
        catch (IOException ex)
        {
            LOGGER.error("Failed to delete generated file ${optSpySource.name}: ${ex.message}")
        }
        return generatedDefinitionFiles
    }

    public List<TFile> generateConfigurationsFor2017AndNewer(String version, String location, IWorkerContext workerContext)
    {
        assert version != null && version.length() > 0 : "Parameter 'version' must not be empty"
        assert location != null && location.length() > 0 : "Parameter 'location' must not be empty"

        List<TFile> generatedDefinitions = new ArrayList<>();

        //C:\Program Files (x86)\Microsoft Visual Studio 2017\Community\VC\Tools\MSVC\14.10.25017\bin\HostX64\x64\cl.exe
        final Pattern pattern = Pattern.compile("Host(.+)");
        TFile compilerRootDir = new TFile(location, "VC/Tools/MSVC");
        List<String> architectures = new ArrayList<>();
        for (TFile compilerDir : compilerRootDir.listFiles(new TDirectoryFileFilter()))
        {
            TFile binDir = new TFile(compilerDir, "bin");
            if (!binDir.exists())
            {
                continue;
            }

            for (TFile hostDir : binDir.listFiles(new TDirectoryFileFilter()))
            {
                final String folderName = hostDir.getName();
                final Matcher matcher = pattern.matcher(folderName);
                if (matcher.find())
                {
                    final String host = matcher.group(1).toLowerCase();
                    for (TFile targetArch : hostDir.listFiles(new TDirectoryFileFilter()))
                    {
                        String targetArchitectureName = targetArch.getName().toLowerCase();
                        if (host.equals(targetArchitectureName))
                        {
                            architectures.add(host);
                        }
                        else
                        {
                            architectures.add(host + "_" + targetArchitectureName);
                        }
                    }
                }
            }
        }

        TFile instDir = new TFile(location);
        TFile binDir = new TFile(instDir, "VC/Auxiliary/Build");

        for (String nextArchitecture : architectures)
        {
            String fileName = generateFileName(version, nextArchitecture)
            if (configFileExists(fileName))
            {
                generatedDefinitions.add(new TFile(m_configDirectory, fileName))
                LOGGER.info("Using previously generated C++ Compiler Definition '${fileName}' in directory '${m_configDirectory.normalizedAbsolutePath}'")
            }
            else if (failedConfigFileExists(fileName))
            {
                LOGGER.info("Skipping generation of previously failed C++ Compiler Definition '" + fileName + "'")
            }
            else
            {
                workerContext.working("Generate definition for Visual Studio C++ Compiler '${version}' '${nextArchitecture}'", true)
                TFile definitionFile = generateConfigurationFor(new TFile(m_configDirectory, generateDefinitionName(version, nextArchitecture)), instDir, binDir, version, nextArchitecture)
                if (definitionFile != null)
                {
                    generatedDefinitions.add(definitionFile)
                }
            }
        }
        return generatedDefinitions;
    }

    private boolean configFileExists(String fileName)
    {
        TFile definitionFile = new TFile(m_configDirectory, fileName)
        return definitionFile.exists()
    }

    private boolean failedConfigFileExists(String fileName)
    {
        TFile definitionFile = new TFile(m_failedConfigDirectory, fileName)
        return definitionFile.exists()
    }

    private void generateFailedFile(String fileName, String errorMsg)
    {
        assert fileName != null && fileName.length() > 0 : "Parameter 'fileName' must not be empty"
        assert errorMsg != null && errorMsg.length() > 0 : "Parameter 'errorMsg' must not be empty"

        if (!m_failedConfigDirectory.exists() || !m_failedConfigDirectory.isDirectory())
        {
            m_failedConfigDirectory.mkdir()
        }

        TFile failedFile = new TFile(m_failedConfigDirectory, fileName)
        try
        {
            FileUtility.writeFile(failedFile.getAbsolutePath(), StandardCharsets.UTF_8, errorMsg, true)
        }
        catch (IOException e)
        {
            LOGGER.error("Error writing failed file information: " + errorMsg, e)
        }
    }

    @Override
    public List<TFile> listGeneratedConfigurations()
    {
        return Arrays.asList(FileUtility.listFilesInDir(m_configDirectory, CompilerDefinitionSearchPath.COMPILER_DEF_FILE_REGEXP))
    }

    @Override
    public String getLatestVersion()
    {
        return m_latestVersion
    }

    @Override
    public String getDefaultVisualCppCompilerDefinition()
    {
        return m_locations == null ? null : "VisualCpp_${m_latestVersion}_${X86}"
    }

    @Override
    public List<TFile> removeGeneratedConfigurations() throws IOException
    {
        TrueZipFacade.clear(m_configDirectory)
        List<TFile> filesToRemove = listGeneratedConfigurations()
        for (TFile file : filesToRemove)
        {
            file.rm()
        }

        if (m_failedConfigDirectory.exists())
        {
            m_failedConfigDirectory.rm_r()
        }
        return filesToRemove
    }

    @Override
    public List<TFile> listFailedConfigurations()
    {
        if (m_failedConfigDirectory.exists())
        {
            return Arrays.asList(m_failedConfigDirectory.listFiles())
        }

        return new ArrayList<TFile>(0)
    }
}